!	A TASM/i8086 backend for T -- Copyright (C) 1997,1998 Nils M. Holm
!	See t.t for details.

const	BPW	=	2;	! bytes per word !
const	PUSHDOWN=	1;	! push-down stack !

const	OBUFL	=	1024;

var	Obuf[OBUFL], Obp;

decl	length(1), newlab(0), strcpy(2);


flush() do
	writes(Obuf);
	Obp := 0;
end


initcg() do
	Obp := 0;
end


endcg() do
	flush();
end


tx(s) do
	var	k;

	k := length(s);
	if (k+Obp >= OBUFL) flush();
	strcpy(@Obuf[Obp], s);
	Obp := Obp + k;
end


pr(s) do
	tx("\t");
	tx(s);
end


nl() tx("\n");

col() tx(":");


totext() do
	pr(".code"); nl();
end


todata() do
	pr(".data"); nl();
end


tobss() do
	pr("ends"); nl();
	tx("_BSS\tsegment word public 'BSS'"); nl();
end


num(n) do
	tx(ntoa(n, 0));
end


lab(n) do
	tx("_L_"); num(n);
end


push() do
	pr("push\tax"); nl();
end


pop() do
	pr("pop\tax"); nl();
end


cg_pop() pop();


pop2() do
	pr("pop\tcx"); nl();
end


cg_extdecl(s) do
	pr("extrn\t"); tx(s); nl();
end


cg_globdecl(s, l, z) do
	tobss();
	pr("even"); nl();
	ie (z) do
		tx(s); tx("\tdw\t"); num(z); tx(" dup(?)"); nl();
	end
	else do
		tx(s); pr("dw\t?"); nl();
	end
	totext();
end


cg_loadglob(s, a, n) do
	pr("mov\tax,");
	if (n) tx("offset ");
	tx(s); nl();
	push();
end


cg_loadlocl(a, n) do
	ie (n) do
		pr("mov\tax,"); num(-a*BPW); nl();
		pr("add\tax,bp"); nl();
	end
	else do
		pr("mov\tax,[bp+"); num(-a*BPW); tx("]"); nl();
	end
	push();
end


cg_indirect() do
	pop();
	pr("mov\tsi,ax"); nl();
	pr("mov\tax,[si]"); nl();
	push();
end


cg_deref(lval) do
	pop2();
	pop();
	pr("shl\tcx,1"); nl();		! BPW = 2 !
	pr("add\tax,cx"); nl();
	if (\lval) do
		pr("mov\tsi,ax"); nl();
		pr("mov\tax,[si]"); nl();
	end
	push();
end


cg_saveglob(s, a) do
	pop();
	pr("mov\t"); tx(s); tx(",ax"); nl();
end


cg_savelocl(a) do
	pop();
	pr("mov\t[bp+"); num(-a*BPW); tx("],ax"); nl();
end


cg_storevec() do
	pop2();
	pop();
	pr("mov\tsi,ax"); nl();
	pr("mov\t[si],cx"); nl();
end


cg_stack(n) do
	if (n) do
		pr("sub\tsp,"); num(n*BPW); nl();
	end
end


cg_unstack(n) do
	if (n) do
		pr("add\tsp,"); num(n*BPW); nl();
	end
end


cg_call(s, a, n) do
	pr("call\tnear ptr "); tx(s); nl();
	cg_unstack(n);
	push();
end


cg_numlit(n) do
	pr("mov\tax,"); num(n); nl();
	push();
end


cg_strlit(s) do
	var	i, l, k;

	k := length(s);
	todata();
	l := newlab();
	lab(l);
	tx("\tdw\t");
	for (i=0, k) do
		if (i /\ \(i mod 10)) do
			nl();
			pr("dw\t");
		end
		num(s[i]);
		if (i mod 10 \= 9) tx(",");
	end
	if (i /\ \(i mod 10)) tx(",");
	tx("0"); nl();
	totext();
	return l;
end


cg_loadlab(l) do
	pr("mov\tax,offset "); lab(l); nl();
	push();
end


cg_sveclit() do
	var	v;

	todata();
	v := newlab();
	lab(v);
	return v;
end


cg_vecmem(n, f) do
	ie (f)  do
		pr("dw\toffset "); lab(n); nl();
	end
	else do
		pr("dw\t"); num(n); nl();
	end
end


cg_eveclit() do
	totext();
end


cg_negate() do
	pop();
	pr("neg\tax"); nl();
	push();
end


cg_lognot() do
	pop();
	pr("neg\tax"); nl();
	pr("sbb\tax,ax"); nl();
	pr("not\tax"); nl();
	push();
end


cg_binnot() do
	pop();
	pr("not\tax"); nl();
	push();
end


op2(s) do
	pop2();
	pop();
	pr(s); tx("\tax,cx"); nl();
	push();
end


cg_multiply() do
	pop2();
	pop();
	pr("imul\tcx"); nl();
	push();
end


cg_divide() do
	pop2();
	pop();
	pr("cwd"); nl();
	pr("idiv\tcx"); nl();
	push();
end


cg_modulo() do
	pop2();
	pop();
	pr("cwd"); nl();
	pr("idiv\tcx"); nl();
	pr("push\tdx"); nl();
end


cg_add() do
	op2("add");
end


cg_subtract() do
	op2("sub");
end



cg_binand() do
	op2("and");
end


cg_binor() do
	op2("or");
end


cg_binxor() do
	op2("xor");
end


cg_binshl() do
	pop2();
	pop();
	tx("shl\tax,cl"); nl();
	push();
end


cg_binshr() do
	pop2();
	pop();
	tx("shr\tax,cl"); nl();
	push();
end


relop(rel) do
	var	y, n;

	y := newlab();
	n := newlab();
	pop2();
	pop();
	pr("cmp\tax,cx"); nl();
	ie (rel = '<')
		pr("jl");
	else ie (rel = '>')
		pr("jg");
	else ie (rel = 'L')
		pr("jle");
	else ie (rel = 'G')
		pr("jge");
	else ie (rel = '=')
		pr("je");
	else ie (rel = '#')
		pr("jne");
	else do
		select(1,2);
		writes("bad relation in relop()\n");
		halt;
	end
	pr(""); lab(y); nl();
	pr("xor\tax,ax"); nl();
	pr("jmp\t"); lab(n); nl();
	lab(y); tx(":\tmov\tax,1"); nl();
	lab(n); col(); nl();
	push();
end


cg_equ() do relop('='); end
cg_nequ() do relop('#'); end
cg_less() do relop('<'); end
cg_grtr() do relop('>'); end
cg_lteq() do relop('L'); end
cg_gteq() do relop('G'); end


cg_incrglob(s, a, v) do
	pr("add\t"); tx(s); tx(","); num(v); nl();
end


cg_incrlocl(a, v) do
	pr("add\tword ptr [bp+"); num(-a*BPW); tx("],"); num(v); nl();
end


cg_tcond(n, flags) do
	var	l;

	l := newlab();
	pop();
	pr("or\tax,ax"); nl();
	ie (flags)
		pr(flags=1-> "je\t": "jne\t");
	else
		pr("jne\t");
 	lab(l); nl();
	if (flags) push();
	pr("jmp\t"); lab(n); nl();
	lab(l); col(); nl();
end


cg_fcond(y, n) do
	pr("jmp\t"); lab(y); nl();
	lab(n); col(); nl();
end


cg_econd(y) do
	lab(y); col(); nl();
end


cg_forcond(step, e) do
	var	l;

	l := newlab();
	pop2();
	pop();
	pr("cmp\tax,cx"); nl();
	pr(step>0-> "jl\t": "jg\t"); lab(l); nl();
	pr("jmp\t"); lab(e); nl();
	lab(l); col(); nl();
end


cg_jump(l) do
	pr("jmp\t"); lab(l); nl();
end


cg_mark(l) do
	lab(l); col(); nl();
end


cg_fnhdr(s, l) do
	tx(s); col(); nl();
	lab(l); col(); nl();
	pr("push\tbp"); nl();
	pr("mov\tbp,sp"); nl();
end


cg_fnexit(r) do
	lab(r); col();
	pr("pop\tbp"); nl();
	pr("ret"); nl();
end


cg_genhalt() do
	pr("xor\tax,ax"); nl();
	push();
	pr("call\tnear ptr __exit"); nl();
end


cg_prologue() do
	initcg();
	tx(";  Compiled by T 3r3t -- Copyright (C) 1997,1998 Nils M. Holm");
	nl();
	tx(";  Code generated for TASM(i8086)"); nl();
	nl();
	pr(".model small"); nl();
	pr("quirks"); nl();
	totext();
	cg_extdecl("__exit");
	cg_extdecl("_LIBINIT");
	tx("DGROUP\tgroup\t_DATA,_BSS"); nl();
	tobss();
	totext();
	pr("mov\tax,DGROUP"); nl();
	pr("mov\tds,ax"); nl();
	pr("mov\tes,ax"); nl();
	pr("mov\tss,ax"); nl();
	pr("call\tnear ptr _LIBINIT"); nl();
	pr("jmp\tSTART"); nl();
end


cg_start() do
	tx("START:");
	pr("mov\tbp,sp"); nl();
end


cg_epilogue() do
	cg_genhalt();
	pr("end"); nl();
	endcg();
end

