!	A SYMBOLIC backend for T -- Copyright (C) 1997 Nils M. Holm
!	See t.t for details.

const	BPW	=	1;	! bytes per word !
const	PUSHDOWN=	1;	! push-down stack !

const	OBUFL	=	1024;

var	Obuf[OBUFL], Obp;

decl	length(1), newlab(0), strcpy(2);


const	ILABEL='L', IPOP='x', IDECL='d', ILDG='g', ILDGV='V', ILDL='p',
	ILDLV='v', IIND='i', IDEREF='X', ISAVG='G', ISAVL='P', ISTORE='S',
	ISTACK='s', IUNSTK='u', ICALL='c', INUM='n', ISTR='$', ILDLAB='l',
	IDATA='D', INEG='_', ILNOT='\\', IBNOT='~', IMUL='*', IDIV='/',
	IMOD='m', IADD='+', ISUB='-', IBAND='&', IBOR='|', IBXOR='^',
	IBSHL='{', IBSHR='}', IEQU='=', INEQU='#', ILESS='<', IGRTR='>',
	ILTEQ='(', IGTEQ=')', IINCG='Z', IINCL='z', IBRF='b', IBRT='B',
	IJUMP='j', IUFOR='f', IDFOR='F', IHDR='r', IEND='R', IHALT='h',
	IINIT='T', IPUSH='!', INOP='.';


flush() do
	writes(Obuf);
	Obp := 0;
end


initcg() Obp := 0;


endcg() flush();


tx(s) do
	var	k;

	k := length(s);
	if (k+Obp >= OBUFL) flush();
	strcpy(@Obuf[Obp], s);
	Obp := Obp + k;
end


pi(i) do
	var	is[2];

	is[0] := i;
	is[1] := 0;
	tx(is);
end


num(n) do
	ie (n<0) do
		tx("%");
		tx(ntoa(-n, 0));
	end
	else do
		tx(ntoa(n, 0));
	end
end


lab(n) do
	pi(ILABEL); num(n);
end


pop() pi(IPOP);


cg_pop() pi(IPOP);


cg_extdecl(s) ;


cg_globdecl(s, l, z) do
	lab(l);
	pi(IDECL); num(z*BPW);
end


cg_loadglob(s, a, n) do
	pi(n-> ILDGV: ILDG);
	num(a);
end


cg_loadlocl(a, n) do
	pi(n-> ILDLV: ILDL);
	num(-a*BPW);
end


cg_indirect() pi(IIND);


cg_deref(lval) do
	pi(IDEREF);
	if (\lval) cg_indirect();
end


cg_saveglob(s, a) do
	pi(ISAVG); num(a);
end


cg_savelocl(a) do
	pi(ISAVL); num(-a*BPW);
end


cg_storevec() pi(ISTORE);


cg_stack(n) do
	if (n) do
		pi(ISTACK); num(n*BPW);
	end
end


cg_unstack(n) do
	if (n) do
		pi(IUNSTK); num(n*BPW);
	end
end


cg_call(s, a, n) do
	pi(ICALL); lab(a); tx(";"); num(n);
end


cg_numlit(n) do
	pi(INUM); num(n);
end


cg_strlit(s) do
	var	i, l, k;

	k := length(s);
	l := newlab();
	lab(l);
	pi(ISTR);
	num(length(s));
	pi(ISTR);
	for (i=0, k) do
		num(s[i]);
		if (i<k-1) tx(";");
	end
	return l;
end


cg_loadlab(l) do
	pi(ILDLAB); lab(l);
end


cg_sveclit() do
	var	v;

	v := newlab();
	lab(v);
	return v;
end


cg_vecmem(n, f) do
	pi(IDATA);
	ie (f)
		lab(n);
	else
		num(n);
end


cg_eveclit() ;


cg_negate() pi(INEG);

cg_lognot() pi(ILNOT);

cg_binnot() pi(IBNOT);

cg_multiply() pi(IMUL);

cg_divide() pi(IDIV);

cg_modulo() pi(IMOD);

cg_add() pi(IADD);

cg_subtract() pi(ISUB);

cg_binand() pi(IBAND);

cg_binor() pi(IBOR);

cg_binxor() pi(IBXOR);

cg_binshl() pi(IBSHL);

cg_binshr() pi(IBSHR);

cg_equ() pi(IEQU);

cg_nequ() pi(INEQU);

cg_less() pi(ILESS);

cg_grtr() pi(IGRTR);

cg_lteq() pi(ILTEQ);

cg_gteq() pi(IGTEQ);


cg_incrglob(s, a, v) do
	pi(IINCG); num(a); tx(";"); num(v);
end


cg_incrlocl(a, v) do
	pi(IINCL); num(-a*BPW); tx(";"); num(v);
end


cg_tcond(n, flags) do
	var	l;

	ie (flags)
		pi(flags=1-> IBRT: IBRF);
	else
		pi(IBRF);
	pi(flags-> IPUSH: INOP);
	pi(IJUMP); lab(n);
end


cg_fcond(y, n) do
	pi(IJUMP); lab(y); lab(n);
end


cg_econd(y) lab(y);


cg_forcond(step, e) do
	var	l;

	l := newlab();
	pi(step>0-> IUFOR: IDFOR); lab(l);
	pi(IJUMP); lab(e);
	lab(l);
end


cg_jump(l) do
	pi(IJUMP); lab(l);
end


cg_mark(l) lab(l);


cg_fnhdr(s, l) do
	lab(l);
	pi(IHDR);
end


cg_fnexit(r) do
	lab(r);
	pi(IEND);
end


cg_genhalt() pi(IHALT);


cg_prologue() do
	initcg();
	pi(IINIT);
	pi(IJUMP);
	lab(0);
end


cg_start() do
	cg_mark(0);
	pi(IHDR);
end


cg_epilogue() do
	cg_genhalt();
	endcg();
end

