!	T 3r3t -- Copyright (C) 1997,1998 Nils M. Holm
!	<fs29@rumms.uni-mannheim.de>
!
!	T is a compiler for a simple procedural language (T).
!	It reads a source program from stdin and writes code
!	described in `c.t' to stdout.
!	The compiler uses an ultra-simple code generator which
!	produces code that simulates a stack machine.
!
!	The T compiler is free software. Do what you want with it,
!	but do not remove any copyright notices.
!
!	The risk of using T is entirely with you, so don't complain.


const	BUFSIZE=	1026,	! Must be <= 2050 !
	SYMBSPACE=	3072,
	NSPACE=		4096,
	TEXTLEN=	129,
	MAXTBL=		128;

const	META=	256;

const	CNST=	1,
	GLOBAL=	2,
	FUNCT=	4,
	PROTO=	8;

const	SNAME=	0,
	SFLAGS=	1,
	SSIZE=	2,
	SVAL=	3,
	SYMLEN=	4;

var	Symbols[SYMBSPACE];
var	St;
var	Sym_equ, Sym_mod, Sym_sub;
var	Names[NSPACE];
var	Nt;
var	Line;
var	Errcount;
var	Token;
var	Text[TEXTLEN];
var	Value;
var	Op;
var	Label;
var	Startlab;
var	Endlab;
var	Exitlab;
var	Looplevl;
var	Locladdr;
var	Buffer[BUFSIZE];
var	Cp, Ep, Nomore;


const	ENDFILE=%1;

const	SYMBOL=0, NUMBER=1, STRING=2;

const	BINOP=10, DISOP=11, CONOP=12, UNOP=20, ADDROP=21, LPAREN=30,
	RPAREN=31, LBRACK=40, RBRACK=41, SEMI=50, COMMA=60, ASSIGN=70,
	COND=80, COLON=81;

const	OPREC=	0,
	OLEN=	1,
	ONAME=	2,
	OTOK=	3,
	OID=	4,
	OPLEN=	5;

var	Ops;

const	KCONST=100, KDECL=101, KDO=102, KELSE=103, KEND=104, KFOR=105,
	KHALT=106, KIE=107, KIF=108, KLEAVE=109, KLOOP=110, KRETURN=111,
	KVAR=112, KWHILE=113;

decl	expr(0), stmt(0), compound(0), declaration(1);


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! CODE GENERATOR !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


initvars() do
	St := 0;
	Nt := 0;
	Line := 1;
	Errcount := 0;
	Label := 1;
	Startlab := 0;
	Endlab := 0;
	Exitlab := 0;
	Looplevl := 0;
	Locladdr := 1;
	Cp := 0;
	Ep := 0;
	Nomore := 0;
	Ops :=	[
		[ 6, 1, "+",	BINOP,	1	],
		[ 7, 2, "*",	BINOP,	2	],
		[ 0, 1, ";",	SEMI,	0	],
		[ 0, 1, ",",	COMMA,	0	],
		[ 0, 1, "(",	LPAREN,	0	],
		[ 0, 1, ")",	RPAREN,	0	],
		[ 0, 1, "[",	LBRACK,	0	],
		[ 0, 1, "]",	RBRACK,	0	],
		[ 3, 1, "=",	BINOP,	3	],
		[ 5, 1, "&",	BINOP,	4	],
		[ 5, 1, "|",	BINOP,	5	],
		[ 5, 1, "^",	BINOP,	6	],
		[ 0, 1, "@",	ADDROP,	0	],
		[ 0, 1, "~",	UNOP,	7	],
		[ 0, 1, ":",	COLON,	0	],
		[ 0, 2, ":=",	ASSIGN,	0	],
		[ 0, 1, "\\",	UNOP,	8	],
		[ 1, 2, "\\/",	DISOP,	0	],
		[ 3, 2, "\\=",	BINOP,	9	],
		[ 4, 1, "<",	BINOP,	10	],
		[ 4, 2, "<=",	BINOP,	11	],
		[ 5, 2, "<<",	BINOP,	12	],
		[ 4, 1, ">",	BINOP,	13	],
		[ 4, 2, ">=",	BINOP,	14	],
		[ 5, 2, ">>",	BINOP,	15	],
		[ 6, 1, "-",	BINOP,	16	],
		[ 0, 2, "->",	COND,	0	],
		[ 7, 2, "/",	BINOP,	17	],
		[ 2, 2, "/\\",	CONOP,	0	],
		[ 7, 1, "mod",	BINOP,	18	],
		[ 0, 0, 0,	0,	0	]
		];
end


strequ(a, b) do
	var	i;

	i := 0;
	while (a[i] /\ b[i]) do
		if (a[i] - b[i]) leave;
		i := i+1;
	end
	return a[i] = b[i];
end


length(a) do
	var	k;

	k := 0;
	while (a[k]) k := k+1;
	return k;
end


strcpy(a, b) do
	var	i;

	i := 0;
	while (b[i]) do
		a[i] := b[i];
		i := i+1;
	end
	a[i] := 0;
end


error(m, s) do
	var	buf[12], o;

	o := select(1, 2);
	writes(ntoa(Line, buf));
	writes(": ");
	writes(m);
	if (s) do
		writes(": ");
		writes(s);
	end
	newline();
	select(1, o);
	Errcount := Errcount +1;
end


fatal(m, s) do
	error(m, s);
	select(1, 2);
	writes("terminating."); newline();
	halt;
end


fillbuf() do
	var	i;

	if (Nomore) return 0;
	for (i=Cp, Ep) Buffer[i-Cp] := Buffer[i];
	i := Ep-Cp;
	Cp := 0;
	Ep := reads(@Buffer[i], BUFSIZE/2-1);
	if (\Ep) Nomore := 1;
	Ep := Ep + i;
end


eof() return Nomore /\ Cp >= Ep;


getce() do
	var	c;

	c := Buffer[Cp]; Cp := Cp+1;
	if (c \= '\\') return c;
	c := Buffer[Cp]; Cp := Cp+1;
	if (c = 'a') return '\a';
	if (c = 'b') return '\b';
	if (c = 'e') return '\e';
	if (c = 'f') return '\f';
	if (c = 'n') return '\n';
	if (c = 'q') return '"' | META;
	if (c = 'r') return '\r';
	if (c = 's') return '\s';
	if (c = 't') return '\t';
	if (c = 'v') return '\v';
	return c;
end


findkw(s)
	ie (s[0] = 'c') do
		if (strequ(s, "const")) return KCONST;
		return 0;
	end
	else ie (s[0] = 'd') do
		if (strequ(s, "decl")) return KDECL;
		if (strequ(s, "do")) return KDO;
		return 0;
	end
	else ie (s[0] = 'e') do
		if (strequ(s, "else")) return KELSE;
		if (strequ(s, "end")) return KEND;
		return 0;
	end
	else ie (s[0] = 'f') do
		if (strequ(s, "for")) return KFOR;
		return 0;
	end
	else ie (s[0] = 'h') do
		if (strequ(s, "halt")) return KHALT;
		return 0;
	end
	else ie (s[0] = 'i') do
		if (strequ(s, "ie")) return KIE;
		if (strequ(s, "if")) return KIF;
		return 0;
	end
	else ie (s[0] = 'l') do
		if (strequ(s, "leave")) return KLEAVE;
		if (strequ(s, "loop")) return KLOOP;
		return 0;
	end
	else ie (s[0] = 'm') do
		if (strequ(s, "mod")) return BINOP;
		return 0;
	end
	else ie (s[0] = 'r') do
		if (strequ(s, "return")) return KRETURN;
		return 0;
	end
	else ie (s[0] = 'v') do
		if (strequ(s, "var")) return KVAR;
		return 0;
	end
	else ie (s[0] = 'w') do
		if (strequ(s, "while")) return KWHILE;
		return 0;
	end
	else do
		return 0;
	end


findop(c) do
	var	i, j;

	i := 0;
	while (Ops[i][OLEN]) do
		if (c = Ops[i][ONAME][0]) do
			ie (Ops[i+1][OLEN] > 1) do
				j := i; i := i+1;
				c := Buffer[Cp]; Cp := Cp+1;
				while (Ops[i][OLEN] = 2) do
					if (c = Ops[i][ONAME][1]) do
						Op := i;
						return Ops[i][OTOK];
					end
					i := i+1;
				end
				Cp := Cp-1;
				Op := j;
				return Ops[j][OTOK];
			end
			else do
				Op := i;
				return Ops[i][OTOK];
			end
		end
		i := i+1;
	end
	return 0;
end


findop2(s) do
	var	i;

	i := 0;
	while (Ops[i][OLEN]) do
		if (strequ(Ops[i][ONAME], s)) return i;
		i := i+1;
	end
	fatal("bad text in findop2()", s);
end


scan() do
	var	c, i, bc[3];

	if (Cp >= Ep-TEXTLEN) fillbuf();
	c := Buffer[Cp]; Cp := Cp+1;
	while (1) do
		while (	c = '\s' \/ c = '\t' \/ c = '\n' \/
			c = '\r' \/ c = '\f'
		) do
			if (c = '\n') Line := Line+1;
			c := Buffer[Cp]; Cp := Cp+1;
		end
		if (c \= '!') leave;
		while (c \= '\n') do
			c := Buffer[Cp]; Cp := Cp+1;
		end
		if (Cp >= Ep-TEXTLEN) fillbuf();
	end
	if (eof()) return ENDFILE;
	if (	'a' <= c /\ c <= 'z' \/
		'A' <= c /\ c <= 'Z' \/
		c = '_'
	) do
		i := 1;
		Text[0] := '_';
		while (1) do
			ie ('A' <= c /\ c <= 'Z')
				c := c-'A'+'a';
			else if (\('a' <= c /\ c <= 'z' \/
				'0' <= c /\ c <= '9' \/
				c = '_')
			)
				leave;
			if (i < TEXTLEN-1) Text[i] := c;
			i := i+1;
			c := Buffer[Cp]; Cp := Cp+1;
		end
		ie (i<TEXTLEN-1)
			Text[i] := 0;
		else
			error("symbol too long", 0);
		Cp := Cp-1;
		c := findkw(@Text[1]);
		if (c) do
			if (c = BINOP) Op := Sym_mod;
			return c;
		end
		Text[i] := '_';
		Text[i+1] := 0;
		return SYMBOL;
	end
	if ('0' <= c /\ c <= '9' \/ c = '%') do
		Value := 0;
		i := 0;
		if (c = '%') do
			i := 1;
			c := Buffer[Cp]; Cp := Cp+1;
		end
		while ('0' <= c /\ c <= '9') do
			Value := Value*10 + (c-'0');
			c := Buffer[Cp]; Cp := Cp+1;
		end
		Cp := Cp-1;
		if (i) Value := -Value;
		return NUMBER;
	end
	if (c = '\'') do
		Value := getce() & ~META;
		c := Buffer[Cp]; Cp := Cp+1;
		if (c \= '\'') error("missing `''", 0);
		return NUMBER;
	end
	if (c = '"') do
		i := 0;
		c := getce();
		while (c \= '"') do
			if (i < TEXTLEN) Text[i] := c & ~META;
			i := i+1;
			c := getce();
			if (eof()) fatal("unexpected EOF", 0);
		end
		ie (i<TEXTLEN)
			Text[i] := 0;
		else
			error("string too long", 0);
		return STRING;
	end
	i := findop(c);
	if (i) return i;
	bc[0] := c/16 + (c/16 > 9-> 'A'-10: '0');
	bc[1] := c mod 16 + (c mod 16 > 9-> 'A'-10: '0');
	bc[2] := 0;
	fatal("bad input character", bc);
end


findsym(name) do
	var	i, s;

	i := St-SYMLEN;
	while (i >= 0) do
		s := Symbols[i+SNAME];
		if (s[1] = name[1] /\ strequ(name, s)) return i;
		i := i-SYMLEN;
	end
	return %1;
end


newsym(name, size, val, flags) do
	var	k, sp;

	k := length(name);
	sp := findsym(name);
	if (name[0] \= '*' /\ sp \= %1) do
		if (flags \= FUNCT \/ Symbols[sp+SFLAGS] \= (FUNCT|PROTO)) do
			error("duplicate symbol", name);
			return sp;
		end
	end
	if (St >= SYMBSPACE) fatal("too many symbols", name);
	if (Nt + k >= NSPACE) fatal("out of name space", name);
	strcpy(@Names[Nt], name);
	Symbols[St+SNAME] := @Names[Nt];
	Nt := Nt+k+1;
	Symbols[St+SSIZE] := size;
	Symbols[St+SVAL] := val;
	Symbols[St+SFLAGS] := flags;
	St := St+SYMLEN;
	return St-SYMLEN;
end


newlab() do
	if (\Label) fatal("out of labels", 0);
	Label := Label+1;
	return Label-1;
end


synch(t) do
	ie (t) do
		while (Token \= t) do
			Token := scan();
			if (Token = ENDFILE)
				fatal("EOF found in error recovery", 0);
		end
	end
	else do
		Token := scan();
	end
end


match(t, m) do
	ie (t = Token)
		Token := scan();
	else
		error(m, 0);
end


xsemi() match(SEMI, "missing `;'");

xlparen() match(LPAREN, "missing `('");

xrparen() match(RPAREN, "missing `)'");

equsign() ie (Token \= BINOP \/ Op \= Sym_equ)
		error("missing `='", 0);
	else
		Token := scan();


constval() do
	var	k, v;

	v := 1;
	ie (Token = NUMBER) do
		v := Value;
		Token := scan();
	end
	else ie (Token = SYMBOL) do
		k := findsym(Text);
		ie (k = %1) do
			error("undefined symbol", Text);
		end
		else do
			ie (\(Symbols[k+SFLAGS] & CNST))
				error("not a constant", Text);
			else
				v := Symbols[k+SVAL];
		end
		Token := scan();
	end
	else do
		error("constant value expected", 0);
	end
	return v;
end


fncall(k) do
	var	n;

	n := 0;
	if (k = %1) error("call of non-function", 0);
	if (k >= 0 /\ \(Symbols[k+SFLAGS] & FUNCT))
		error("call of non-function", Symbols[k+SNAME]);
	xlparen();
	if (Token \= RPAREN) while (1) do
		expr();
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	if (k >= 0 /\ Symbols[k+SSIZE] \= n)
		error("wrong number of arguments", Symbols[k+SNAME]);
	if (k >= 0) cg_call(Symbols[k+SNAME], Symbols[k+SVAL], n);
	xrparen();
end


loadsym(k) do
	if (Symbols[k+SFLAGS] & FUNCT) return 0;
	ie (Symbols[k+SFLAGS] & GLOBAL)
		cg_loadglob(Symbols[k+SNAME], Symbols[k+SVAL],
			Symbols[k+SSIZE]);
	else
		cg_loadlocl(Symbols[k+SVAL], Symbols[k+SSIZE]);
end


savesym(k) ie (Symbols[k+SFLAGS] & GLOBAL)
		cg_saveglob(Symbols[k+SNAME], Symbols[k+SVAL]);
	else
		cg_savelocl(Symbols[k+SVAL]);


genop(id) do
	if (id = 1) return cg_add();
	if (id = 2) return cg_multiply();
	if (id = 3) return cg_equ();
	if (id = 4) return cg_binand();
	if (id = 5) return cg_binor();
	if (id = 6) return cg_binxor();
	if (id = 7) return cg_binnot();
	if (id = 8) return cg_lognot();
	if (id = 9) return cg_nequ();
	if (id = 10) return cg_less();
	if (id = 11) return cg_lteq();
	if (id = 12) return cg_binshl();
	if (id = 13) return cg_grtr();
	if (id = 14) return cg_gteq();
	if (id = 15) return cg_binshr();
	if (id = 16) return cg_subtract();
	if (id = 17) return cg_divide();
	if (id = 18) return cg_modulo();
end


address(lval) do
	var	k;

	if (Token \= SYMBOL) do
		error("symbol expected", 0);
		return %1;
	end
	k := findsym(Text);
	if (k = %1) error("undefined symbol", Text);
	ie (k \= %1 /\ (Symbols[k+SFLAGS] & CNST)) do
		cg_numlit(Symbols[k+SVAL]);
	end
	else do
		if (\lval /\ k \= %1) loadsym(k);
	end
	Token := scan();
	while (Token = LBRACK) do
		if (k >= 0 /\ (Symbols[k+SFLAGS] & (CNST|FUNCT)))
			error("bad subscript", Symbols[k+SNAME]);
		if (lval) do
			ie (k = %1)
				cg_indirect();
			else
				loadsym(k);
		end
		Token := scan();
		expr();
		cg_deref(lval);
		match(RBRACK, "missing `]'");
		k := %1;
	end
	return k;
end


table() do
	var	t[MAXTBL], tf[MAXTBL], p, v, i;

	p := 0;
	Token := scan();
	while (	Token = SYMBOL \/ Token = NUMBER \/
		Token = LBRACK \/ Token = STRING
	) do
		ie (Token = SYMBOL \/ Token = NUMBER) do
			t[p] := constval();
			tf[p] := 0;
		end
		else ie (Token = STRING) do
			t[p] := cg_strlit(Text);
			tf[p] := 1;
			Token := scan();
		end
		else do
			t[p] := table();
			tf[p] := 2;
		end
		if (p >= MAXTBL) fatal("table too big", 0);
		p := p+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	v := cg_sveclit();
	for (i=0, p) cg_vecmem(t[i], tf[i]);
	cg_eveclit();
	match(RBRACK, "missing `]'");
	return v;
end


factor() do
	var	o, k;

	ie (Token = NUMBER) do
		cg_numlit(Value);
		Token := scan();
	end
	else ie (Token = SYMBOL) do
		k := address(0);
		ie (Token = LPAREN) do
			fncall(k);
		end
		else do
			if (k >= 0 /\ (Symbols[k+SFLAGS] & FUNCT))
				error("incomplete function call",
					Symbols[k+SNAME]);
		end
	end
	else ie (Token = STRING) do
		cg_loadlab(cg_strlit(Text));
		Token := scan();
	end
	else ie (Token = LBRACK) do
		cg_loadlab(table());
	end
	else ie (Token = ADDROP) do
		Token := scan();
		k := address(1);
		if (k >= 0) do
			ie (Symbols[k+SFLAGS] & CNST)
				error("cannot take @const", Symbols[k+SNAME]);
			else ie (Symbols[k+SFLAGS] & FUNCT)
				error("cannot take @function",
					Symbols[k+SNAME]);
			else ie (Symbols[k+SFLAGS] & GLOBAL)
				cg_loadglob(Symbols[k+SNAME],
					Symbols[k+SVAL], 1);
			else
				cg_loadlocl(Symbols[k+SVAL], 1);
		end
	end
	else ie (Token = BINOP) do
		if (Op \= Sym_sub) error("bad unary operator", 0);
		Token := scan();
		factor();
		cg_negate();
	end
	else ie (Token = UNOP) do
		o := Op;
		Token := scan();
		factor();
		genop(Ops[o][OID]);
	end
	else ie (Token = LPAREN) do
		Token := scan();
		expr();
		xrparen();
	end
	else do
		error("bad expression", 0);
		synch(0);
	end
end


emitop(stk, sp) do
	var	o;

	sp := sp-1;
	genop(Ops[stk[sp]][OID]);
	return sp;
end


binary() do
	var	stk[5], sp, fac;

	fac := 0;
	sp := 0;
	factor();
	while (1) do
		ie (Token = BINOP) do
			if (fac) do
				error("missing factor", 0);
				sp := sp-1;
			end
			while (sp /\ Ops[Op][OPREC] <= Ops[stk[sp-1]][OPREC])
				sp := emitop(stk, sp);
			if (sp >= 5) fatal("binary(): stack overflow", 0);
			stk[sp] := Op; sp := sp+1;
			fac := 1;
			Token := scan();
		end
		else do
			ie (fac) do
				factor();
				fac := 0;
			end
			else do
				while (sp) sp := emitop(stk, sp);
				return 0;
			end
		end
	end
end


conj() do
	var	e;

	e := 0;
	binary();
	if (Token = CONOP) e := newlab();
	while (Token = CONOP) do
		Token := scan();
		cg_tcond(e, 2);
		binary();
	end
	if (e) cg_econd(e);
end


disj() do
	var	e;

	e := 0;
	conj();
	if (Token = DISOP) e := newlab();
	while (Token = DISOP) do
		Token := scan();
		cg_tcond(e, 1);
		conj();
	end
	if (e) cg_econd(e);
end


expr() do
	var	y, n;

	y := newlab();
	n := newlab();
	disj();
	if (Token = COND) do
		Token := scan();
		cg_tcond(n, 0);
		expr();
		cg_fcond(y, n);
		match(COLON, "missing `:'");
		expr();
		cg_econd(y);
	end
end


ie_stmt(alt) do
	var	y, n;

	n := newlab();
	if (alt) y := newlab();
	Token := scan();
	xlparen();
	expr();
	xrparen();
	cg_tcond(n, 0);
	stmt();
	if (alt) do
		cg_fcond(y, n);
		match(KELSE, "missing `ELSE'");
		stmt();
	end
	cg_econd(alt-> y: n);
end


while_stmt() do
	var	s, e, levl;

	s := Startlab;
	e := Endlab;
	levl := Looplevl;
	Startlab := newlab();
	Endlab := newlab();
	Looplevl := Locladdr;
	Token := scan();
	xlparen();
	cg_mark(Startlab);
	expr();
	cg_tcond(Endlab, 0);
	xrparen();
	stmt();
	cg_jump(Startlab);
	cg_econd(Endlab);
	Startlab := s;
	Endlab := e;
	Looplevl := levl;
end


for_stmt() do
	var	k, step, r, s, e, levl;

	step := 1;
	s := Startlab;
	e := Endlab;
	levl := Looplevl;
	Startlab := newlab();
	r := newlab();
	Endlab := newlab();
	Looplevl := Locladdr;
	Token := scan();
	xlparen();
	if (Token = SYMBOL) do
		k := findsym(Text);
		if (k < 0) error("undefined symbol", Text);
		if (k \= %1 /\ Symbols[k+SFLAGS] & (CNST|FUNCT))
			error("bad lvalue", Text);
		if (k \= %1 /\ Symbols[k+SSIZE])
			error("index may not be a vector", Text);
	end
	match(SYMBOL, "symbol name expected");
	equsign();
	expr();
	if (k \= %1) savesym(k);
	match(COMMA, "missing `,'");
	cg_mark(r);
	if (k \= %1) loadsym(k);
	expr();
	if (Token = COMMA) do
		Token := scan();
		step := constval();
	end
	xrparen();
	cg_forcond(step, Endlab);
	stmt();
	cg_mark(Startlab);
	ie (Symbols[k+SFLAGS] & GLOBAL)
		cg_incrglob(Symbols[k+SNAME], Symbols[k+SVAL], step);
	else
		cg_incrlocl(Symbols[k+SVAL], step);
	cg_jump(r);
	cg_mark(Endlab);
	Startlab := s;
	Endlab := e;
	Looplevl := levl;
end


asg_or_call() do
	var	k;

	if (Token \= SYMBOL) do
		error("bad statement", 0);
		synch(SEMI);
		return 0;
	end
	k := address(1);
	ie (Token = ASSIGN) do
		Token := scan();
		expr();
		ie (k \= %1) do
			if (Symbols[k+SSIZE])
				error("missing subscript", Symbols[k+SNAME]);
			if (Symbols[k+SFLAGS] & (CNST|FUNCT))
				error("bad lvalue", Symbols[k+SNAME]);
			savesym(k);
		end
		else do
			cg_storevec();
		end
	end
	else ie (Token = LPAREN) do
		fncall(k);
		cg_pop();
	end
	else do
		error("bad statement", 0);
		synch(SEMI);
	end
end


stmt() do
	ie (Token = KDO) do
		compound();
	end
	else ie (Token = KIE) do
		ie_stmt(1);
	end
	else ie (Token = KIF) do
		ie_stmt(0);
	end
	else ie (Token = KWHILE) do
		while_stmt();
	end
	else ie (Token = KFOR) do
		for_stmt();
	end
	else ie (Token = KLEAVE) do
		Token := scan();
		xsemi();
		if (\Endlab) error("`LEAVE' in wrong context", 0);
		cg_unstack(Locladdr - Looplevl);
		cg_jump(Endlab);
	end
	else ie (Token = KLOOP) do
		Token := scan();
		xsemi();
		if (\Startlab) error("`LOOP' in wrong context", 0);
		cg_unstack(Locladdr - Looplevl);
		cg_jump(Startlab);
	end
	else ie (Token = KRETURN) do
		Token := scan();
		expr();
		xsemi();
		if (\Exitlab) error("`RETURN' in wrong context", 0);
		cg_pop();
		cg_unstack(Locladdr-1);
		cg_jump(Exitlab);
	end
	else ie (Token = KHALT) do
		Token := scan();
		xsemi();
		cg_genhalt();
	end
	else ie (Token = SEMI) do
		Token := scan();
	end
	else ie (Token = KELSE) do
		Token := scan();
		error("`ELSE' with no matching `IE'", 0);
	end
	else do
		asg_or_call();
		xsemi();
	end
end


cleanup() do
	var	k, la;

	la := Locladdr;
	k := findsym("*");
	if (k \= %1) do
		St := k;
		Nt := Symbols[k+SSIZE];
		Locladdr := Symbols[k+SVAL];
	end
	cg_unstack(la - Locladdr);
end


compound() do
	var	ot;

	ot := Locladdr;
	newsym("*", Nt, Locladdr, 0);
	Token := scan();
	while (Token = KVAR \/ Token = KCONST) declaration(1);
	cg_stack(Locladdr - ot);
	while (Token \= KEND) do
		if (eof()) fatal("unexpected EOF", 0);
		stmt();
	end
	Token := scan();
	cleanup();
end


adjust(k, n) do
	n := n+3;
	while (k < St) do
		Symbols[k+SVAL] := Symbols[k+SVAL]-n;
		k := k+SYMLEN;
	end
end


fndecl() do
	var	p, k, lv, n, addr;

	n := 0;
	addr := 2;
	p := findsym(Text);
	Exitlab := newlab();
	k := newsym(Text, 0, 0, FUNCT);
	lv := newsym("*", Nt, 1, 0)+SYMLEN;
	Token := scan();
	xlparen();
	if (Token \= RPAREN) while (1) do
		if (Token = SYMBOL) do
			newsym(Text, 0, addr, 0);
			addr := addr+1;
		end
		match(SYMBOL, "argument name expected");
		n := n+1;
		if (Token \= COMMA) leave;
		Token := scan();
	end
	Symbols[k+SSIZE] := n;
	ie (p \= %1 /\ (Symbols[p+SFLAGS] & PROTO)) do
		if (Symbols[p+SSIZE] \= n)
			error("argument count mismatch", Symbols[k+SNAME]);
		Symbols[k+SVAL] := Symbols[p+SVAL];
	end
	else do
		Symbols[k+SVAL] := newlab();
	end
	cg_fnhdr(Symbols[k+SNAME], Symbols[k+SVAL]);
	adjust(lv, n);
	xrparen();
	stmt();
	cleanup();
	cg_fnexit(Exitlab);
	Exitlab := 0;
end


fnproto() do
	var	k;

	Token := scan();
	while (1) do
		if (Token \= SYMBOL) do
			error("symbol expected", 0);
			synch(SEMI); Token := scan();
			return 0;
		end
		k := newsym(Text, 0, 0, FUNCT|PROTO);
		Token := scan();
		xlparen();
		Symbols[k+SSIZE] := constval();
		Symbols[k+SVAL] := newlab();
		xrparen();
		if (Token \= COMMA) leave;
		Token := scan();
	end
	xsemi();
end


declaration(local) do
	var	t, k;

	ie (Token = KVAR \/ Token = KCONST) do
		t := Token;
		Token := scan();
		while (1) do
			if (Token \= SYMBOL) do
				error("symbol expected", 0);
				synch(SEMI); Token := scan();
				return 0;
			end
			k := newsym(Text, 0, 0,
				(t=KCONST-> CNST: 0) | (local-> 0 : GLOBAL));
			Token := scan();
			ie (t = KCONST) do
				equsign();
				Symbols[k+SVAL] := constval();
			end
			else do
				if (Token = LBRACK) do
					Token := scan();
					Symbols[k+SSIZE] := constval();
					match(RBRACK, "missing ']'");
					if (Symbols[k+SSIZE] < 1)
						error("bad array size",
							Symbols[k+SNAME]);
				end
				ie (local) do
					Symbols[k+SVAL] := Locladdr;
					if (PUSHDOWN /\ Symbols[k+SSIZE])
						Symbols[k+SVAL] :=
							Symbols[k+SVAL] +
							Symbols[k+SSIZE]-1;
					Locladdr := Locladdr +
						(Symbols[k+SSIZE]->
						Symbols[k+SSIZE]: 1);
				end
				else do
					Symbols[k+SVAL] := newlab();
					cg_globdecl(Symbols[k+SNAME],
						Symbols[k+SVAL],
						Symbols[k+SSIZE]);
				end
			end
			if (Token \= COMMA) leave;
			Token := scan();
		end
		xsemi();
	end
	else ie (Token = KDECL) do
		fnproto();
	end
	else ie (Token = SYMBOL) do
		fndecl();
	end
	else do
		error("bad declaration", 0);
		synch(0);
	end
end


external(s, n) do
	newsym(s, n, newlab(), FUNCT);
	cg_extdecl(s);
end


initfuncs() do
	external("_open_", 2);
	external("_close_", 1);
	external("_erase_", 1);
	external("_select_", 2);
	external("_reads_", 2);
	external("_writes_", 1);
	external("_newline_", 0);
	external("_aton_", 1);
	external("_ntoa_", 2);
	external("_pack_", 2);
	external("_unpack_", 2);
	newsym("__version__", 0, 32, CNST|GLOBAL);
	Sym_equ := findop2("=");
	Sym_sub := findop2("-");
	Sym_mod := findop2("mod");
end


do
	initvars();
	cg_prologue();
	initfuncs();
	Token := scan();
	while (Token = KVAR \/ Token = KCONST \/ Token = SYMBOL \/
		Token = KDECL
	)
		declaration(0);
	if (Token \= KDO) fatal("`DO' or declaration expected", 0);
	cg_start();
	compound();
	cg_epilogue();
end

